/* 
 * Copyright (C) 2011  Xin Zhang and Franck van Breugel
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can find a copy of the GNU General Public License at
 * <http://www.gnu.org/licenses/>.
 */

package probabilistic.listener;

import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.vm.ChoiceGenerator;
import gov.nasa.jpf.search.Search;
import gov.nasa.jpf.search.SearchListener;
import probabilistic.vm.Probabilistic;

/**
 * This listener prints the states and the transitions visited by the
 * search in the following formats: 
 * 0 - 0.5 -> 1, denoting that there is a transition from state 0 to 
 * state 1 with probability 0.5, and
 * 0 -------> 1, denoting that there is a transition from state 0 to
 * state 1.
 *
 * @author Xin Zhang
 * @author Franck van Breugel
 */
public class StateSpacePrinter extends ListenerAdapter implements SearchListener 
{
    private int source; // id of the source of the next transition

    /**
     * Initializes the listener.
     */
    public StateSpacePrinter() 
    {
	source = -1; // -1 represents undefined
    }

    /**
     * Executed whenever the search advances to the next state.
     *
     * @param search object that provides information about the search.
     */
    public void stateAdvanced(Search search) 
    {
	int target = search.getStateId();
	if (source != -1) 
	{
	    ChoiceGenerator cg = search.getVM().getChoiceGenerator();
	    if (cg instanceof Probabilistic) 
	    {
		double probability = ((Probabilistic) cg).getProbability();
		System.out.printf("%d -%.2f-> %d%n", source, probability, target);
	    } 
	    else 
	    {
		System.out.printf("%d ------> %d%n", source, target);
	    }
	}
	source = target;
    }

    /**
     * Executed whenever the search backtracks.
     *
     * @param search object that provides information about the search.
     */
    public void stateBacktracked(Search search)
    {
	source = search.getStateId();
    }

    /**
     * Executed whenever the search moves to a previously visited state.
     *
     * @param search object that provides information about the search.
     */
    public void stateRestored(Search search) 
    {
	source = search.getStateId();
    }
}
