/* 
 * Copyright (C) 2013  Franck van Breugel
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can find a copy of the GNU General Public License at
 * <http://www.gnu.org/licenses/>.
 */

package probabilistic.listener;

import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.vm.ChoiceGenerator;
import gov.nasa.jpf.search.Search;
import gov.nasa.jpf.search.SearchListener;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import probabilistic.vm.Probabilistic;

/**
 * This listener serializes the states and transitions visited if JPF hits
 * its memory limit.  
 *
 * @author Franck van Breugel
 */
public class StateSpaceSerializer extends ListenerAdapter implements SearchListener 
{
    private List<Transition> transitions; // transitions searched
    private Set<Integer> endStates; // end states
    private int source; // id of the source of the next transition

    /**
     * Initializes the listener.
     */
    public StateSpaceSerializer() 
    {
	transitions = new ArrayList<Transition>();
	endStates = new HashSet<Integer>();
	source = -1; // -1 represents undefined
    }

    /**
     * Executed whenever the search advances to the next state.
     *
     * @param search object that provides information about the search.
     */
    public void stateAdvanced(Search search) 
    {
	int target = search.getStateId();
	if (search.isEndState())
	{
	    endStates.add(target);
	}
	if (source != -1) 
	{
	    ChoiceGenerator cg = search.getVM().getChoiceGenerator();
	    if (cg instanceof Probabilistic) 
	    {
		double probability = ((Probabilistic) cg).getProbability();
		transitions.add(new Transition(source, target, probability));
	    } 
	    else 
	    {
		transitions.add(new Transition(source, target, 1.0));
	    }
	}
	source = target;
    }

    /**
     * Executed whenever the search backtracks.
     *
     * @param search object that provides information about the search.
     */
    public void stateBacktracked(Search search)
    {
	source = search.getStateId();
    }

    /**
     * Executed whenever the search moves to a previously visited state.
     *
     * @param search object that provides information about the search.
     */
    public void stateRestored(Search search) 
    {
	source = search.getStateId();
    }

    /*
    public void searchFinished(Search search)
    {
	for (Integer state : endStates) 
	{
	    transitions.add(new Transition(state, state, 1.0));
	}

	try
	{
	    String name = "t" + System.currentTimeMillis() + ".ser";
	    FileOutputStream file = new FileOutputStream(name);
	    ObjectOutputStream out = new ObjectOutputStream(file);
	    out.writeObject(transitions);
	    out.close();
	    file.close();
	} 
	catch (IOException e)
	{
	    e.printStackTrace();
	}
    }
    */

    /**
     * Executed whenever the search hits some constraint.
     *
     * @param search object that provides information about the search.
     */
    public void searchConstraintHit(Search search)
    {
	for (Integer state : endStates) 
	{
	    transitions.add(new Transition(state, state, 1.0));
	}
	
	try
	{
	    String name = "/tmp/t" + System.currentTimeMillis() + ".ser";
	    FileOutputStream file = new FileOutputStream(name);
	    ObjectOutputStream out = new ObjectOutputStream(file);
	    out.writeObject(transitions);
	    out.close();
	    file.close();
	    
	    System.out.println("The underlying Markov chain has been serialized in the file " + name);
	} 
	catch (IOException e)
	{
	    e.printStackTrace();
	}
    }
}
