/* 
 * Copyright (C) 2011  Xin Zhang and Franck van Breugel
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can find a copy of the GNU General Public License at
 * <http://www.gnu.org/licenses/>.
 */

package probabilistic.tool;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.HashMap;
import java.util.TreeSet;

import probabilistic.listener.Transition;

/**
 * Given the name of a file consisting of a serialized list of transitions, this 
 * app compute the progress and prints it.
 *
 * @author Xin Zhang
 * @author Franck van Breugel
 */
public class Progress
{
    private Progress() {}

    /**
     * The app prints the progress. 
     *
     * @param name the name of the file containing the serialized list of transitions.
     */
    public static void main(String[] args) 
    {
	if (args.length == 0) 
	{
	    System.out.println("Provide the name of the file containing the serialized list of transitions as command line argument");
	} 
	else 
	{
	    String name = args[0];
	    try
	    {
		// read list of transitions from file
		FileInputStream file = new FileInputStream(name);
		ObjectInputStream in = new ObjectInputStream(file);
		List<Transition> transitions = (List<Transition>) in.readObject();
		in.close();
		file.close();

		// maximal state
		int max = Integer.MIN_VALUE;

		// mapping from states to their sets of predecessors
		HashMap<Integer, TreeSet<Integer>> predecessors = new HashMap<Integer, TreeSet<Integer>>();
		
		// mapping from states to probabilities of outgoing transitions
		HashMap<Integer, ArrayList<Double>> probabilities = new HashMap<Integer, ArrayList<Double>>();
		
		for (Transition transition : transitions)
		{
		    int source = transition.getSource();
		    int target = transition.getTarget();
		    double probability = transition.getProbability();

		    // update max
		    max = Math.max(max, Math.max(source, target));

		    // update predecessors
		    if (!predecessors.containsKey(target))
		    {
			predecessors.put(target, new TreeSet<Integer>());
		    }
		    predecessors.get(target).add(source);

		    // update probabilities
		    if (!probabilities.containsKey(source))
		    {
			probabilities.put(source, new ArrayList<Double>());
		    }
		    probabilities.get(source).add(probability);
		    if (!probabilities.containsKey(target))
		    {
			probabilities.put(target, new ArrayList<Double>());
		    }
		}

		// add transition for not fully explored states to sink state (max + 1)
		HashMap<Integer, Double> notFullyExplored = new HashMap<Integer, Double>();
		for (int state = 0; state <= max; state++)
		{
		    if (probabilities.containsKey(state))
		    {
			Collections.sort(probabilities.get(state));
			double sum = 0.0;
			for (double probability : probabilities.get(state))
			{
			    sum += probability;
			}
			if (sum < 1.0)
			{
			    notFullyExplored.put(state, sum);
			}
		    }
		}
		predecessors.put(max + 1, new TreeSet<Integer>(notFullyExplored.keySet()));

		// find states that cannot reach sink
		TreeSet<Integer> temp = new TreeSet<Integer>();
		temp.add(max + 1);
		TreeSet<Integer> canReachSink = new TreeSet<Integer>();
		while (!temp.isEmpty())
		{
		    int state = temp.first();
		    temp.remove(state);
		    canReachSink.add(state);
		    if (predecessors.containsKey(state))
		    {
			for (Integer predecessor : predecessors.get(state))
			{
			    if (!canReachSink.contains(predecessor))
			    {
				temp.add(predecessor);
			    }
			}
		    }
		}
		temp = null;
		TreeSet<Integer> cannotReachSink = new TreeSet<Integer>();
		for (int state = 0; state <= max; state++)
		{
		    if (probabilities.containsKey(state) && !canReachSink.contains(state))
		    {
			cannotReachSink.add(state);
		    }
		}
		canReachSink = null;

		if (cannotReachSink.contains(0))
		{
		    System.out.printf("%.10f%n", 0.0);
		}
		else
		{
		    // find states that always reach sink
		    temp = new TreeSet<Integer>();
		    temp.addAll(cannotReachSink);
		    TreeSet<Integer> notAlwaysReachSink = new TreeSet<Integer>();
		    while (!temp.isEmpty())
		    {
			int state = temp.first();
			temp.remove(state);
			notAlwaysReachSink.add(state);
			if (predecessors.containsKey(state))
			{
			    for (Integer predecessor : predecessors.get(state))
			    {
				if (!notAlwaysReachSink.contains(predecessor))
			        {
				    temp.add(predecessor);
				}
			    }
			}
		    }
		    temp = null;
		    TreeSet<Integer> alwaysReachSink = new TreeSet<Integer>();
		    for (int state = 0; state <= max; state++)
		    {
			if (probabilities.containsKey(state) && !notAlwaysReachSink.contains(state))
			{
			    alwaysReachSink.add(state);
			}
		    }

		    if (alwaysReachSink.contains(0))
		    {
			System.out.printf("%.10f%n", 1.0);
		    }
		    else
		    {
			// find remaining states (those can reach the sink state but do not always reach the sink state)
			ArrayList<Integer> remaining = new ArrayList<Integer>();
			for (int state = 0; state <= max; state++)
			{
			    if (probabilities.containsKey(state) && !cannotReachSink.contains(state) && !alwaysReachSink.contains(state))
			    {
				remaining.add(state);
			    }
			}

			// matrix for Jacobi
			final int size = remaining.size();
			double[][] matrix = new double[size][size];
			for (Transition transition : transitions)
			{
			    int source = transition.getSource();
			    int target = transition.getTarget();
			    double probability = transition.getProbability();

			    if (remaining.contains(source) && remaining.contains(target))
			    {
				matrix[remaining.indexOf(source)][remaining.indexOf(target)] += probability;
			    }
			}
			for (int row = 0; row < size; row++)
			{
			    for (int column = 0; column < size; column++)
			    {
				if (row == column)
				{
				    matrix[row][column] = 1.0 - matrix[row][column];
				}
				else
				{
				    matrix[row][column] = -matrix[row][column];
				}
			    }
			}

			// vector for Jacobi
			double[] vector = new double[size];
			for (Transition transition : transitions)
			{
			    int source = transition.getSource();
			    int target = transition.getTarget();
			    double probability = transition.getProbability();

			    if (remaining.contains(source) && alwaysReachSink.contains(target))
			    {
				vector[remaining.indexOf(source)] += probability;
			    }
			}
			for (Integer state : remaining)
			{
			    if (notFullyExplored.containsKey(state))
			    {
				vector[remaining.indexOf(state)] += 1.0 - notFullyExplored.get(state);
			    }
			}

			// Jacobi
			double[] diagonal = new double[size];
			for (int i = 0; i < size; i++)
			{
			    diagonal[i] = 1.0 / matrix[i][i];
			}
			double[] current = new double[size];
			double difference;
			do
			{
			    double[] previous = current;
			    current = new double[size];
			    for (int i = 0; i < size; i++)
			    {
				double[] product = new double[size - 1];
				for (int j = 0, h = 0; j < size; j++)
				{
				    if (i != j)
				    {
					product[h] = matrix[i][j] * previous[j];
					h++;
				    }
				}
				Arrays.sort(product);
				double sum = 0.0;
				for (int h = 0; h < size - 1; h++)
				{
				    sum += product[h];
				}
				current[i] = diagonal[i] * (vector[i] - sum);
			    }
			    difference = 0.0;
			    for (int i = 0; i < size; i++)
			    {
				difference = Math.max(difference, Math.abs(previous[i] - current[i]));
			    }
			}
			while (difference > 0.00000000001);

			System.out.printf("%.10f%n", current[0]);
		    }
		}
	    }
	    catch (IOException e)
	    {
		System.out.println("Failed to read the file " + name);
		e.printStackTrace();
	    }
	    catch (ClassNotFoundException e)
	    {
		System.out.println("Failed to convert the content of the file " + name);
		e.printStackTrace();
	    }
	}
    }
}
